;;; Directory Local Variables
;;; For more information see (info "(emacs) Directory Variables")

((nil
  (bug-reference-bug-regexp . "#\\(?2:[0-9]+\\)")
  (bug-reference-url-format . "https://github.com/postgres-plr/plr/issues/%s"))
 (c-mode
  (c-basic-offset . 4)
  (c-file-style . "bsd")
  (fill-column . 78)
  (indent-tabs-mode . t)
  (tab-width . 4))
 (ess-r-mode
  (ess-style . RStudio))
 (sql-mode
  (sql-indent-offset . 2)
  (fill-column . 78)
  (indent-tabs-mode)))
